--[[
-- RPG Engine (c) 2018 Ale32bit
]]--

local Engine = {
  debug = false,
  assetsPathBase = "assets/",
  tiles = {},
  specialTiles = {},
  items = {},
  entities = {},
  npcs = {},
  map = {},
  player = {
    side = "front",
  },
  scale = 4,
  noclip = false,
}

Engine.texturePathBase = Engine.assetsPathBase.."textures/"
Engine.mapsPathBase = Engine.assetsPathBase.."maps/"

local config = {}

function Engine.log(...)
  print("[ENGINE]",...)
end

function Engine.load()
  Engine.log("Loading Engine")
  config.screenWidth = love.graphics.getWidth()
  config.screenHeight = love.graphics.getHeight()
  Engine.canvas = love.graphics.newCanvas(config.screenWidth, config.screenHeight)
  love.graphics.setCanvas(Engine.canvas)
  love.graphics.clear()
  love.graphics.setCanvas()
  Engine.log("Size",config.screenWidth, config.screenWidth)
end

function Engine.addTile(name, texture)
  local image = love.graphics.newImage(Engine.texturePathBase.."tiles/"..texture..".png")
  table.insert(Engine.tiles, {name=name, texture=image})
  Engine.log("[NEW TILE]",#Engine.tiles.." "..name)
  return #Engine.tiles
end

function Engine.addItem(name, texture)
  local image = love.graphics.newImage(Engine.texturePathBase.."items/"..texture..".png")
  table.insert(Engine.items, {name=name, texture=image})
  Engine.log("[NEW ITEM]",#Engine.items.." "..name)
  return #Engine.items
end

function Engine.setSpecialTile(name,path)
  local image = love.graphics.newImage(path)
  Engine.specialTiles[name] = image
  Engine.log("[NEW TILE]",name)
end

function Engine.setPlayer(path, speed, options)
  Engine.player.sides = {
    front = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/front.png"),
    back  = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/back.png"),
    left  = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/leftSide.png"),
    right = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/rightSide.png"),
  }
  options.focus = options.focus or {}
  options.hitbox = options.hitbox or {}
  Engine.player.focus = {
    x = options.focus.x or 0,
    y = options.focus.y or 0,
  }
  Engine.player.hitbox = {
    x = options.hitbox.x or 0,
    y = options.hitbox.y or 0,
    w = options.hitbox.w or 0,
    h = options.hitbox.h or 0,
  }
  Engine.player.speed = speed or 96
  Engine.log("Player set!")
end

function Engine.addEntity(name, options)
  local data = {
    name = name,
  }
  local entity = {
    name = data.name,
  }
  
  table.insert(Engine.entities,{
      name = name or "???",
      entity = entity,
      data = data,
    })
  
  Engine.log("[NEW ENTITY]",name,#Engine.entities)
  return entity, #Engine.entities
end

function Engine.getEntity(id)
  return Engine.entities[id].entity
end

function Engine.summon(id,x,y)
  x=x or 1
  y=y or 1
  table.insert(Engine.map.entities,{id=id,x=x,y=y})
end

function Engine.addNPC(name,path,options)
  local sides
  if path then
    sides = {
      front = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/front.png"),
      back  = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/back.png"),
      left  = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/leftSide.png"),
      right = love.graphics.newImage(Engine.texturePathBase.."entities/"..path.."/rightSide.png"),
    }
  end
  options.hitbox = options.hitbox or {}
  options.focus = options.focus or {}
  local focus = {
    x = options.focus.x or 0,
    y = options.focus.y or 0,
  }
  local hitbox = {
    x = options.hitbox.x or 0,
    y = options.hitbox.y or 0,
    w = options.hitbox.w or 0,
    h = options.hitbox.h or 0,
  }
  options.color = options.color or {}
  table.insert(Engine.npcs, {
    texture = sides or false,
    name = name or "n/a",
    focus = focus,
    hitbox = hitbox,
    color = {
      r = options.color[1] or 255,
      g = options.color[2] or 255,
      b = options.color[3] or 255,
      a = options.color[4] or 255,
    }
  })
  return #Engine.npcs
end

function Engine.summonNPC(id,x,y)
  x=x or Engine.player.x
  y=y or Engine.player.y
  table.insert(Engine.map.npcs,{id=id,x=x,y=y,facing="front",hitbox=Engine.npcs[id].hitbox,focus=Engine.npcs[id].focus})
  local nID = #Engine.map.npcs
  return {
    setPos = function(x,y)
      Engine.map.npcs[nID].x = x
      Engine.map.npcs[nID].y = y
    end,
    getPos = function()
      return Engine.map.npcs[nID].x, Engine.map.npcs[nID].y
    end,
    setFacing = function(side)
      Engine.map.npcs[nID].facing = side
    end,
    getFacing = function()
      return Engine.map.npcs[nID].facing
    end,
  }
end

function Engine.setMap(map)
  local old = Engine.map
  Engine.map = map
  Engine.player.x = map.config.player.x or 0
  Engine.player.y = map.config.player.y or 0
  Engine.player.side = map.config.player.facing or "front"
  return old
end

function Engine.loadMap(name)
  local map = require(Engine.mapsPathBase..name)
  map.entities = map.entities or {}
  map.npcs = map.npcs or {}
  return map
end

function Engine.resize(w,h)
  config.screenWidth = w
  config.screenHeight = h
  Engine.log("Window Resize",w,h)
end

function Engine.draw()
  
  love.graphics.setCanvas(Engine.canvas)
  love.graphics.clear()
  
  local screen_width = love.graphics.getWidth() / Engine.scale
  local screen_height = love.graphics.getHeight() / Engine.scale
  
  local tx = Engine.player.x - screen_width / 2
  local ty = Engine.player.y - screen_height / 2
  
  love.graphics.setColor(255,255,255,255)
  --love.graphics.translate(-tx, -ty)
  local offsetx, offsety = 0,0
  local map = Engine.map
  local mapConfig = map.config
  
  -- Draw map
  for _, layer in pairs(map.maps.view) do 
    for _,y in pairs(layer) do
      for _,x in pairs(y) do
        if x > 0 then
          local image
          if Engine.tiles[x] then
            image = Engine.tiles[x].texture
          else
            image = Engine.specialTiles["unknown"]
          end
          love.graphics.draw(image,offsetx-tx,offsety-ty)
        end
        offsetx = offsetx + mapConfig.tileSize[1]
      end
      offsetx = 0
      offsety = offsety+mapConfig.tileSize[2]
    end
    
    offsetx = 0
    offsety = 0
    
  end
  
  -- draw all map npcs
  for id, npc in pairs(Engine.map.npcs) do
    local npcdata = Engine.npcs[npc.id]
    if npcdata.texture then
      love.graphics.setColor(npcdata.color.r,npcdata.color.g,npcdata.color.b,npcdata.color.a)
      love.graphics.draw(npcdata.texture[npc.facing],npc.x-tx+npcdata.focus.x,npc.y-ty+npcdata.focus.y)
    end
    
    -- debug
    if Engine.debug then --debug focus point
      love.graphics.setPointSize(5)
      love.graphics.setColor(255,0,0)
      love.graphics.points(npc.x-tx, npc.y-ty) -- ITS NOT A PEEPEE
      
      --draw hitbox
      love.graphics.rectangle("line",
        npc.x +npcdata.focus.x -tx + npcdata.hitbox.x,
        npc.y +npcdata.focus.y -ty + npcdata.hitbox.y,
        npcdata.hitbox.w,
        npcdata.hitbox.h
      )
    end
  end
  
  love.graphics.setColor(255,255,255,255) --set white color, full alpha
  love.graphics.draw(Engine.player.sides[Engine.player.side],(screen_width / 2)+Engine.player.focus.x,(screen_height / 2)+Engine.player.focus.y) -- draw player
  
  if Engine.debug then --draw debug focus point
    love.graphics.setPointSize(5)
    love.graphics.setColor(255,0,0)
    love.graphics.points((screen_width / 2), (screen_height / 2)) -- ITS NOT A PEEPEE
    
    local player = Engine.player
    -- draw player hitbox-
    love.graphics.rectangle("line",
      player.x + player.focus.x -tx + player.hitbox.x,
      player.y + player.focus.y -ty + player.hitbox.y,
      player.hitbox.w,
      player.hitbox.h
    )
    
    local offsetx, offsety = 0,0
    
    love.graphics.setLineWidth(0.5)
    
    for _,y in pairs(map.maps.walls) do
      for _,x in pairs(y) do
        if x == 1 then
          love.graphics.rectangle("line",offsetx-tx,offsety-ty,mapConfig.tileSize[1],mapConfig.tileSize[2])
        end
        offsetx = offsetx + mapConfig.tileSize[1]
      end
      offsetx = 0
      offsety = offsety+mapConfig.tileSize[2]
    end
    
    love.graphics.setColor(255,255,255,255)
  end
  
  love.graphics.setCanvas()
  love.graphics.draw(Engine.canvas,0,0,0,Engine.scale)
  
  love.graphics.print("Pass: "..tostring(Engine.canPass()),1,1)
  love.graphics.print("Wall: "..tostring(Engine.isWall(Engine.player.x, Engine.player.y)),1,16)
  love.graphics.print("HitBox: "..tostring(Engine.inHitbox()),1,32)
  
  if Engine.noclip then
    love.graphics.print("Noclip ON",1,love.graphics.getHeight())
  end
end

function Engine.isWall(x,y,map)
  map = map or Engine.map
  local walls = map.maps.walls
  local tileSize = map.config.tileSize
  
  x = math.ceil(x/tileSize[1])
  y = math.ceil(y/tileSize[2])
  
  if walls[y] and walls[y][x] and walls[y][x] == 1 then
    return true
  end
  return false
end

function Engine.inHitboxSing(entA, entB)
  
  local screen_width = love.graphics.getWidth() / Engine.scale
  local screen_height = love.graphics.getHeight() / Engine.scale
  
  local tx = Engine.player.x - screen_width / 2
  local ty = Engine.player.y - screen_height / 2
  
  local a = {}
  
  a.x = entA.x
  a.y = entA.y
  
  a.fx = entA.focus.x
  a.fy = entA.focus.y
  
  a.hx = a.x + a.fx + entA.hitbox.x 
  a.hy = a.y + a.fy + entA.hitbox.y
  a.hw = entA.hitbox.w
  a.hh = entA.hitbox.h
  
  local b = {}
  
  b.x = entB.x
  b.y = entB.y
  
  b.fx = entB.focus.x
  b.fy = entB.focus.y
  
  b.hx = b.x + b.fx + entB.hitbox.x 
  b.hy = b.y + b.fy + entB.hitbox.y
  b.hw = entB.hitbox.w
  b.hh = entB.hitbox.h
  
  -- npc.x + npcdata.focus.x -tx + npcdata.hitbox.x,
  -- npc.y + npcdata.focus.y -ty + npcdata.hitbox.y,
  -- npcdata.hitbox.w,
  -- npcdata.hitbox.h
  
  return a.hx < b.hx+b.hw and
         b.hx < a.hx+a.hw and
         a.hy < b.hy+b.hh and
         b.hy < a.hy+a.hh
end

function Engine.inHitbox(entA)
  entity = entity or Engine.player
  
  for _,npc in pairs(Engine.map.npcs) do
    if Engine.inHitboxSing(entity, npc) then
      return true
    end
  end
  return false
end

function Engine.canPass(entity)
  entity = entity or Engine.player
  
  for _,npc in pairs(Engine.map.npcs) do
    if Engine.inHitboxSing(entity, npc) then
      return false
    end
  end
  
  if Engine.isWall(entity.x, entity.y) then
    return false
  end
  return true
end

function Engine.update(dt)
  local oldX = Engine.player.x
  local oldY = Engine.player.y
  if love.keyboard.isDown("a") or love.keyboard.isDown("left") then
    Engine.player.x = Engine.player.x - Engine.player.speed * dt
    if not Engine.canPass() and not Engine.noclip then
      Engine.player.x = oldX
    end
    Engine.player.side = "left"
  end
  if love.keyboard.isDown("d") or love.keyboard.isDown("right") then
      Engine.player.x = Engine.player.x + Engine.player.speed * dt
      if not Engine.canPass() and not Engine.noclip then
        Engine.player.x = oldX
      end
      Engine.player.side = "right"
  end
  if love.keyboard.isDown("w") or love.keyboard.isDown("up") then
    Engine.player.y = Engine.player.y - Engine.player.speed * dt
    if not Engine.canPass() and not Engine.noclip then
      Engine.player.y = oldY
    end
    Engine.player.side = "back"
  end
  if love.keyboard.isDown("s") or love.keyboard.isDown("down") then
    Engine.player.y = Engine.player.y + Engine.player.speed * dt
    if not Engine.canPass() and not Engine.noclip then
      Engine.player.y = oldY
    end
    Engine.player.side = "front"
  end
end

return Engine